/* 
 * Modbus Library
 * $Id: modbus.c 18 2004-12-08 10:28:05Z rsc $
 */

#include <stdlib.h>
#include <string.h>
#include <sys/types.h>
#include <sys/socket.h>
#include <netinet/in.h>
#include <arpa/inet.h>

#include "modbus.h"

static int transaction_id = 1;

/* Header for Modbus/TCP */
struct modbus_mbap_header {
	uint16_t	transaction_identifier;
	uint16_t	protocol_identifier;
	uint16_t	length;
	uint8_t		unit_identifier;
} __attribute__ ((packed));	
typedef struct modbus_mbap_header modbus_mbap_header_t;

/* Protocol Data Unit */
	
/* Function Code 0x01: Request Coils */
struct req_r_coils { 
	uint8_t  function_code;	
	uint16_t starting_addr; 
	uint16_t quantity;
} __attribute__ ((packed));

struct resp_r_coils {
	uint8_t  function_code;
	uint8_t  byte_count;
} __attribute__ ((packed));
struct error_r_coils {
	uint8_t  function_code;
	uint8_t  exception_code;
} __attribute__ ((packed));

/* Function Code 0x02: Read Discrete Inputs */
struct req_r_discrete_in {
	uint8_t  function_code;
	uint16_t starting_addr;
	uint16_t quantity;
} __attribute ((packed));
struct resp_r_discrete_in {
	uint8_t  function_code;
	uint8_t  byte_count;
	uint8_t  data;
} __attribute__ ((packed));
struct error_r_discrete_in {
	uint8_t  function_code;
	uint8_t  exception_code;
} __attribute__ ((packed));
	
/* Function Code 0x05: Write Single Coil */
struct req_w_single_coil { 
	uint8_t  function_code; 
	uint16_t output_addr; 
	uint16_t output_value; 
} __attribute__ ((packed));
struct resp_w_single_coil {
	uint8_t  function_code; 
	uint16_t output_addr; 
	uint16_t output_value;
} __attribute__ ((packed));
struct error_w_single_coil {
	uint8_t  function_code; 
	uint8_t  exception_code; 
} __attribute__ ((packed));
	
#define modbus_pdu_size 253


int modbus_bind (modbus_dev_t *dev, char *ip, unsigned int port)
{
	int ret;
	
	if ((dev->in_socket_fd = socket (AF_INET, SOCK_STREAM, 0)) < 0) {
		printf("%s: socket() error", __FUNCTION__);
		exit(1);
	}

	dev->in_address.sin_family = AF_INET;
	dev->in_address.sin_addr.s_addr = inet_addr(ip);
	dev->in_address.sin_port = htons (port);

	ret = connect (dev->in_socket_fd, (struct sockaddr*)&dev->in_address, sizeof (dev->in_address));
	if (ret < 0) {
		printf ("%s: could not connect socket\n", __FUNCTION__);
	}
	return ret;
		
}


int modbus_unbind (modbus_dev_t *dev)
{
	close (dev->in_socket_fd); 
}


int modbus_w_single_coil (modbus_dev_t *dev, int coil, int value)
{
	int id;
	uint8_t *buf;
	modbus_mbap_header_t *header;
	struct req_w_single_coil *req;
	struct resp_w_single_coil *resp;
	int ret = 0;
	int packet_size;
	int i;

	id = transaction_id++;

	buf = malloc (sizeof (modbus_mbap_header_t) + modbus_pdu_size);
	header = (modbus_mbap_header_t *)buf;
	req = (struct req_w_single_coil *)(buf + sizeof (modbus_mbap_header_t));

	req->function_code = FKT_W_SINGLE_COIL;
	req->output_addr   = htons (coil);
	req->output_value  = htons (value ? 0xFF00 : 0x0000);

	header->transaction_identifier = htons (id);
	header->protocol_identifier = htons (0x0000); 
	header->length = htons (5); 
	header->unit_identifier = htons (0x0000);
	
	printf ("\n");
	printf ("req    = %p\n", req);
	printf ("header = %p\n", header);
	printf ("buf    = %p\n", buf);

	packet_size = sizeof (modbus_mbap_header_t) + ntohs(header->length);
	buf = realloc (buf, packet_size);

	for (i=0; i<packet_size; i++) {
		printf ("%02x ", buf[i]);
	}
	
	write (dev->in_socket_fd, buf, packet_size);
	
	read (dev->in_socket_fd, buf, 2);
	resp = (struct resp_w_single_coil *) (buf + sizeof (*header));
	
	i = ntohs (resp->function_code);
	if (i & 0x80) {
		printf ("received error: fkt_code=%02x\n", i);
		ret = -1; goto out;
	}

	if ((i & 0x7F) != FKT_W_SINGLE_COIL) {
		printf ("response_fkt != received_fkt: fkt_code=%02x\n", i);
		ret = -1; goto out;
	}

	printf ("received: output-address = %04x\n", resp->output_addr);
	printf ("received: output-value   = %04x\n", resp->output_value);
	
out:	free (buf);
	
	return ret;
}
